use crate::Direction::{Left, Right};
use std::collections::HashMap;

const FILE_STR: &str = include_str!("input.txt");
const START_LOC: &str = "AAA";
const END_LOC: &str = "ZZZ";

#[derive(Debug, Copy, Clone)]
enum Direction {
    Left,
    Right,
}

impl From<char> for Direction {
    fn from(value: char) -> Self {
        match value {
            'L' => Left,
            'R' => Right,
            _ => panic!("Invalid direction char"),
        }
    }
}

type DirectionMap = HashMap<String, (String, String)>;

#[derive(Debug)]
struct Network {
    direction_list: Vec<Direction>,
    map: DirectionMap,
}

impl Network {
    fn parse(input: &str) -> Self {
        let mut lines = input.split('\n');
        let direction_list: Vec<Direction> = lines
            .next()
            .unwrap()
            .trim()
            .chars()
            .map(Direction::from)
            .collect();

        let map = lines
            .skip(1)
            .filter(|l| !l.is_empty())
            .map(|l| {
                let parts: Vec<_> = l.split(" = ").collect();
                let key = parts[0].trim();

                let raw_dirs = parts[1].replace(['(', ')'], "");
                let dirs: Vec<_> = raw_dirs.split(", ").collect();
                let left = dirs[0].trim();
                let right = dirs[1].trim();

                (String::from(key), (String::from(left), String::from(right)))
            })
            .collect::<DirectionMap>();

        Network {
            direction_list,
            map,
        }
    }

    fn count_steps(&self) -> usize {
        let mut count = 0usize;
        let mut next_key = START_LOC;
        let mut current = self.map.get(next_key).unwrap();

        loop {
            for x in self.direction_list.iter() {
                count += 1;
                next_key = match x {
                    Direction::Left => &current.0,
                    Direction::Right => &current.1,
                };

                if next_key == END_LOC {
                    return count;
                }

                current = self
                    .map
                    .get(next_key)
                    .unwrap_or_else(|| panic!("Unable to find key {} in map", next_key));
            }
        }
    }
}

fn part_one() {
    let network = Network::parse(FILE_STR);
    let step_count = network.count_steps();

    println!("Part 1: Step count from AAA to ZZZ {}", step_count);
}

fn main() {
    part_one();
}

#[cfg(test)]
mod tests {
    const EXAMPLE_FILE_STR: &str = include_str!("example-input.txt");
    use super::*;

    #[test]
    fn first_example() {
        let network = Network::parse(EXAMPLE_FILE_STR);
        assert_eq!(6, network.count_steps());
    }
}
